# encoding
CHARSET = r"utf-8"

# path and name
CUT_RESULT_FILE_NAME = r"cut_result.json"
REPORT_FILE_NAME = r"report.html"

# report
BACKGROUND_COLOR = r"#fffaf4"
UNSTABLE_FLAG = r"-1"
UNKNOWN_STAGE_FLAG = r"-2"
IGNORE_FLAG = r"-3"

# config
DEFAULT_THRESHOLD = 0.98

# ext
FFMPEG = r"ffmpeg"
